"Python3 program to perform evolutionary clustering."

##  CUES :
##
##  Clustering Using Evolutionary Search (in Python3) :
##  Copyright 2014, R.S. Forsyth.
##  first version : 18/09/2014
##  last revision : 01/02/2015


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math
import  os, sys
import  random, time
##  import  turtle


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  request
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from gridlib import  vecdist
from gridlib import  copygrid, distest
from py3seal import  Genepool


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("filextn",''),
    ("ntrials",131072,256,1048576),  ##  number of evolutionary trials
    ##  ("randseed",3433,1,999999999),  ##  random seed
    ("jobname","cues"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("datfile",''),  ##  file for input of data
    ("scaling",1,0,2),  ## whether to rescale vals b4 distance calx.
    ("longtime",3600,600,86400),  ##  elapsed secs after which to skip posttidy
    ("distmode","city"),  ##  distance mode
    ("optimode",1,1,3),  ##  selector of fitness function
    ##  ("vars2use",''),  ##  variables to be used
    ("skipvars",''),  ##  variable to be excluded
    ("listfile",''),  ##  output file mainly for debugging
    ("dumpfile",''),  ##  output file for derived data
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"cues.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "1.7"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 4
Dumptail = "_dump"
Listtail = "_list"
Paratail = "_pars"
Popsize = 233
##  character labels for each cluster :
Charlabs = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789*!"
Maxgmax = len(Charlabs)
Ruspini = [1]*20 + [2]*23 + [3]*17 + [4]*15


##  Main program :
##
def main():
    "Main line."
    global Ruspini
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    ##  prepare list file :
    if paradat.listfile:  ##  must be open b4 init()
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listname,'w',encoding="utf8")
    ##  initialization :
    paradat = init(paradat)
    nd = len(paradat.dvec)
    print(nd,"distances.")
    paradat.dvec.sort()
    ##  print(paradat.dmin,paradat.dmid,paradat.basedist,paradat.dmax)
    ##  print(paradat.dvec)

    print(paradat.pals[0],file=paradat.listfile)
    print(paradat.comrades[0],file=paradat.listfile)
    print("number of pals =",len(paradat.pals[0]))
    palztot = 0
    for t in range(paradat.nt):
        palztot += len(paradat.comrades[t])
    print("mean number of comrades =",round(palztot / paradat.nt,4))

    ##  main process :
    ##  establish operational parameter settings :
    fitfunx = [fitness1,fitness2,fitness3]
    ff = fitfunx[int(paradat.optimode)-1]
    paradat.gmax = int(paradat.nt**0.5+1) + int(math.log(paradat.nt+1))
    paradat.gmax = min(Maxgmax,paradat.gmax)  ##  upper limit enforced
    paradat.glen = paradat.rows
    print(paradat.glen,paradat.gmax)
    mu = 1.0
    pool = Genepool("cues1501",ff,paradat,paradat.glen,paradat.gmax,mu,1)
    mutavec = [1.0,Gold,0.5,1.0/Gold]
    paradat.indices = indices = list(range(paradat.rows))
    paradat.ncycles = ncycles = int(math.log1p(paradat.rows))
    if paradat.rows >= 42:
        paradat.subsize = int(math.log1p(paradat.rows)*math.sqrt(paradat.rows))
    elif paradat.rows >= 26:
        paradat.subsize = paradat.rows - 1
    else:
        paradat.subsize = paradat.rows
    ##  evolutionary optimization loop :
    print("cycles,subsize =",ncycles,paradat.subsize)
    bestyet = -Huge ; himp = 0
    for loop in range(ncycles):
        paradat.included = set(random.sample(indices,paradat.subsize))
        paradat.inclist = sorted(paradat.included)  ##  also keep as list
        paradat.reps = paradat.ntrials // paradat.ncycles
        if loop == (ncycles-1):  ##  last cycle; use extras if any left
            reps = paradat.ntrials - paradat.reps*(ncycles-1)
            mp = himp  ##  reuse best mutation rate (pointer)
        else:
            reps = paradat.reps
            mp = loop % len(mutavec)
        print("loop",loop+1,reps,"trials")
        newpals(paradat)  ##  trim neighbour sets
        ##  apply evolutionary optimizer to subset :
        pool.mu = mutavec[mp]  ##  pick mutation rate
        bestval = evclust(paradat,pool,reps,ff)
        print("loop",loop+1,"local score =",round(bestval,7))
        ##  allocate unselected items to a suitable group :
        ##  bi = allocate(paradat,pool)
        bi = inserter(paradat,pool)
        ##  also evaluate on full genestring :
        ##  temp = paradat.included.copy()
        paradat.included = set(paradat.indices)  ##  ensure no exclusions
        paradat.inclist = sorted(paradat.indices)  ##  & keep as list
        oldpals(paradat)  ##  restore neighbour sets
        bv = ff(bi,paradat)  ##  best value
        ##  paradat.included = temp
        nicelist,nicelabs = eugenic(bi,Charlabs)  ##  condensed version
        nc = max(nicelist)+1
        print(nc,"groups.")
        print("loop",loop+1,"global score =",round(bv,7))
        if bv > bestyet:
            bestyet = bv ; bestclus = nicelist[:]
            bestmute = mutavec[mp]  ##  pool.mu
            himp = mp  ##  pointer
            bestloop = loop
            if paradat.glen < 256:
                print(''.join(nicelabs))
            else:
                print(''.join(nicelabs)[:256],"....")

    paradat.included = set(paradat.indices)  ##  ensure all cases evaluated
    paradat.inclist = sorted(paradat.included)
    ##  special chunk for ruspini testing (to go eventually?) :
    if ("ruspini" in paradat.datfile) and paradat.rows == len(Ruspini):
        print(Ruspini)
        for i,f in enumerate(fitfunx):
            r = Ruspini
            print(i,f.__name__)
            print(f(r,paradat))
            r = Ruspini[:45] + [5,5,5] + Ruspini[48:]
            print(f(r,paradat))
                
    ##  decide on final groupings :
    ##  print("eugenic:",bestclus)
    print(max(bestclus)+1,"groups.")
    ##  print("bestmute =",bestmute,bestloop)
    elapsed = time.time() - startime
    print(round(elapsed,4),"secs elapsed.")
    ##  cpusecs = time.clock() - clockon
    ##  print(round(cpusecs,4),"cpu.secs")
    ##  here do final tidying.
    if elapsed < paradat.longtime:  ##  still time to do post-tidying
        newbest,newscore = posttidy(paradat,bestclus,bestyet,ff)
    else:  ##  too long already !
        newbest = bestclus ; newscore = bestyet  ##  just do fast tidying
    ##  newbest1,newval1 = reunite(paradat,bestclus,bestyet,ff)
    newbest2,newval2 = reunite(paradat,newbest,newscore,ff)
    for f in []:  ##  formerly fitfunx:
        print(f.__name__)
        q = f(bestclus,paradat)
        print(round(q,7))
        q = f(newbest,paradat)
        print(round(q,7))
        ##  q = f(newbest1,paradat)
        ##  print(round(q,7))
        q = f(newbest2,paradat)
        print(round(q,7))
    ##  print(paradat.dmid,paradat.dbar)
    ##  print(newbest2)
    nc = len(set(newbest2))
    print(nc,"clusters.")
    difs = 0
    for j,c in enumerate(bestclus):
        cx = newbest2[j]
        if c != cx:
            ##  print(j,c,cx)
            difs += 1
    if difs == 1:
        print("1 change made during post-tidying.")
    else:
        print(difs,"changes made during post-tidying.")
    if difs > paradat.gmax:  ##  paradat.rows ** 0.5:
        print("Probable failure to converge.")
        if paradat.ntrials <= 524288:
            print("Longer run might help:")
            nt = int(paradat.ntrials*Gold)
            print("e.g., ntrials increased to",nt)
        else:
            print("Perhaps no clear cluster structure in data?")
    newbest,nb = eugenic(newbest2,Charlabs)  ##  might need re-tidying.
    print(newbest,file=paradat.listfile)
    print(''.join(nb),file=paradat.listfile)
    
    ##  prepare o/p files :
    if paradat.outfile:
        fo = open(paradat.outfile,'w',encoding="utf-8")
    else:
        outname = paradat.outpath + paradat.jobname + Paratail + ".txt"
        fo = open(outname,'w',encoding="utf-8")
    if paradat.dumpfile:
        dumpname = paradat.dumpfile
    else:
        dumpname = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    if hasattr(paradat,"datamat"):  ##  only overwrite if input given
        dumpfile = open(dumpname,'w',encoding="utf8")

    ##  o/p of results :
    paradat.listfile.close()
    parashow(paradat,fo) ; fo.close()
    if hasattr(paradat,"datamat"):
        dumpout(dumpfile,paradat,newbest)  ##  ,bestclus)
        fullname = os.path.abspath(dumpfile.name)
        print("results dumped on ",fullname)
        dumpfile.close()
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    ##  print(round(time.clock()-clockon,4),"cpu.secs")
    return  0



##  NEWPALS :
def newpals (paradat):
    "trims friends & foes sets to reflect inclusions."

    paradat.oldpals = [0]*paradat.rows ; paradat.oldfoes = [0]*paradat.rows
    paradat.oldcoms = [0]*paradat.rows ; paradat.oldcasts = [0]*paradat.rows
    ##  paradat.nfriends = [1]*paradat.rows
    for i in range(paradat.rows):
        ##  first save full versions :
        paradat.oldpals[i] = paradat.pals[i].copy()
        paradat.oldfoes[i] = paradat.foes[i].copy()
        paradat.oldcoms[i] = paradat.comrades[i].copy()
        paradat.oldcasts[i] = paradat.outcasts[i].copy()
        ##  next restrict to included cases :
        paradat.pals[i] = paradat.pals[i] & paradat.included
        paradat.foes[i] = paradat.foes[i] & paradat.included
        paradat.comrades[i] = paradat.comrades[i] & paradat.included
        paradat.outcasts[i] = paradat.outcasts[i] & paradat.included
        paradat.nfriends[i] = max(len(paradat.pals[i]),len(paradat.foes[i]),1)
        if len(paradat.comrades[i]) < 1:
            paradat.comrades[i].add(paradat.nearest[i])
        if len(paradat.outcasts[i]) < 1:
            paradat.outcasts[i].add(paradat.furthest[i])
        
    return


##  OLDPALS :
def oldpals (paradat):
    "restores friends & foes sets to 'factory settings'."

    for i in range(paradat.rows):
        paradat.pals[i] = paradat.oldpals[i]
        paradat.foes[i] = paradat.oldfoes[i]
        paradat.comrades[i] = paradat.oldcoms[i]
        paradat.outcasts[i] = paradat.oldcasts[i]
        paradat.nfriends[i] = paradat.friends1

    return


##  EVCLUST :
def evclust (paradat,pool,reps,ff):
    "evolutionary clustering with fitness ff on included items."

    bias = []
    pool.initpop(Popsize,bias)
    pool.flea(reps)
    print(round(pool.best,7))
    ##  print(paradat.glen,paradat.gmax)

    return  pool.best  ##  highest fitness score


##  ALLOCATE (removed) :


##  INSERTER :
def inserter (paradat,pool):
    "puts each unselected instance into a suitable cluster."

    gvec = pool.bestitem[:]
    ##  compute cluster sizes :
    sizetab = {}
    for i in paradat.included:
        c = gvec[i]
        if c in sizetab:
            sizetab[c] += 1
        else:  ##  first encounter
            sizetab[c] = 1
    ##  allocate to closest cluster :
    for j in range(paradat.glen):
        if j in paradat.included:
            continue  ##  only consider excluded cases
        dtot = {}  ##  total dist to items in each cluster
        for i,c in enumerate(gvec):
            if i in paradat.included:
                pair = (i,j)
                if c in dtot:
                    dtot[c] += paradat.ddic[pair]
                else:  ##  first time
                    dtot[c] = paradat.ddic[pair]
        ##  next find cluster with least mean distance (median later?) :
        dmin = Huge
        for c in dtot:
            dbar = dtot[c] / sizetab[c]
            if dbar < dmin:
                dmin = dbar ; clus = c
        gvec[j] = clus

    return  gvec


##  FITNESS1 :
def fitness1 (gvec,paradat):
    "sociological fitness function."

    ##  assemble cluster info :
    clustab = {}
    for g in paradat.inclist:
        c = gvec[g]
        if c in clustab:
            clustab[c].add(g)
        else:  ##  first encounter
            clustab[c] = set([g])
    ##  score social security of each item :
    vtot = 0.0
    for item in paradat.inclist:
        c = gvec[item]
        cset = clustab[c]
        ##  fixed size :
        good = cset & paradat.pals[item]  ##  inner
        poor = cset & paradat.foes[item]
        vtot += (len(good) - len(poor)) / paradat.nfriends[item]  ##  paradat.friends1
        ##  variable size (never empty) :
        good = cset & paradat.comrades[item]
        poor = cset & paradat.outcasts[item]
        vtot += (len(good)/len(paradat.comrades[item]))
        vtot -= (len(poor)/len(paradat.outcasts[item]))
        
    return  vtot / (2*len(paradat.included))  ##  mean felicity score


##  FITNESS2 :
def fitness2 (gvec,paradat):
    "fitness function for robust evolutionary clustering."

    gain = cost = 0
    ##  k = paradat.glen*(paradat.glen-1) // 2  ##  comparisons
    k = len(paradat.included)*len(paradat.included) / Gold
    for i in range(paradat.glen):
        if i in paradat.included:
            for j in range(i+1,paradat.glen):
                if not (j in paradat.included):
                    continue
                pair = (i,j)
                dm = paradat.distomed[pair]
                ##  dm = paradat.ddic[pair] - paradat.dmid
                if gvec[i] == gvec[j]:  ##  same group
                    cost += dm  ##  dist is a cost, less is better
                else:
                    gain += dm  ##  large dist is good

    return  (gain-cost) / k


##  FITNESS3 :
def fitness3 (gvec,paradat):
    "alternative fitness function for robust evolutionary clustering."

    inclust = outclust = 0.0
    insides = outsides = 0
    for i in paradat.inclist:
        for j in range(i+1,paradat.glen):
            if not (j in paradat.included):
                continue
            pair = (i,j)
            ##  dm = paradat.distomed[pair]
            dm = paradat.ddic[pair]  ##  exponent applied in init()
            if gvec[i] == gvec[j]:  ##  same group
                inclust += dm  ##  dist is a cost, less is better
                insides += 1
            else:
                outclust += dm
                outsides += 1

    return  (outclust/max(1,outsides) - inclust/max(1,insides)) / paradat.basedist
    ##  return  (outclust / max(1,outsides) - inclust/max(1,insides))


##  MISFITS :
def misfits (gvec,paradat):
    "routine to find items least central to their respective groups."


    fitlist = []
    for i in paradat.inclist:
        inclust = outclust = 0.0
        insides = outsides = 0
        for j in paradat.inclist:
            pair = (i,j)
            ##  dm = paradat.distomed[pair]
            dm = paradat.ddic[pair]  ##  exponent applied in init()
            if gvec[i] == gvec[j]:  ##  same group
                inclust += dm  ##  dist is a cost, less is better
                insides += 1
            else:
                outclust += dm
                outsides += 1
        di = inclust / insides
        do = outclust / outsides
        fitlist.append((do-di*Gold,i))

    return  sorted(fitlist)  ##  most misfitting at front


##  EUGENIC :
def eugenic (bi,charlabs):
    "makes nice printable version of best individual, bi."

    gset = set(bi)
    ##  nc = len(gset)
    catlist = sorted(gset)  ##  ordered list of values used
    nicelist = [] ; lablist = []
    for g in bi:
        ci = catlist.index(g)
        nicelist.append(ci)
        lablist.append(charlabs[ci])

    return  nicelist, lablist


##  NEARFAR :
def nearfar (paradat):
    "returns lists of nearest & furthest neighbours in ddic."

    nearest = [0]*paradat.nt ; furthest = [0]*paradat.nt
    for t in range(paradat.nt):
        n = Huge ; f = -Huge
        for tt in range(paradat.nt):
            if t == tt:
                continue  ##  don't compare with self
            d = paradat.ddic[(t,tt)]
            if d > f:
                f = d ; furthest[t] = tt
            if d < n:
                n = d ; nearest[t] = tt

    return  nearest, furthest


##  GETFILE :
def getfile (paradat):
    "obtains data grid from file."

    paradat.datamat,paradat.taglist,paradat.tagdict = matread(paradat.datfile,
                                                              paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print("column names :")
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    paradat.goodlist = sorted(paradat.goodcols)
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2

    return  0  ##  success, paradat altered


##  PICKPALS :
def pickpals (paradat):
    "picks subset of near & distant cases for each item."

    paradat.friends1 = int(paradat.nt**0.5+0.5)  ##  n of foes is same
    paradat.friends2 = paradat.friends1
    if paradat.nt >= 42:
        paradat.friends2 = int(paradat.nt**0.5*Gold+0.5)
    paradat.pals = [] ; paradat.foes = []  ##  inner
    paradat.comrades = [] ; paradat.outcasts = []  ##  outer
    ndis = len(paradat.dvec)
    palval2 = paradat.dvec[ndis//4]  ##  already sorted
    foeval2 = paradat.dvec[ndis//2]  ##  median inter-dist
    for item in range(paradat.nt):  ##  nt should have been computed
        sortlist = []
        for j in range(paradat.nt):
            d = paradat.ddic[(item,j)]  ##  ddic should have been filled
            if j != item:  ##  don't include self
                sortlist.append((d,j))
        sortlist.sort() ; ns = len(sortlist)
        ##  near ones :
        paradat.pals.append(set())
        paradat.comrades.append(set([paradat.nearest[item]]))
        ##  always include nearest
        np = max(ns//4,paradat.friends2)
        for d,j in sortlist[:np]:  ##  near ones
            if d < palval2:
                paradat.comrades[item].add(j)
                ##  cd have variable number of items
            if len(paradat.pals[item]) < paradat.friends1:
                paradat.pals[item].add(j)  ##  fixed size (same as foes)
        ##  distant ones :
        paradat.foes.append(set()) ; paradat.outcasts.append(set())
        for i in range(ns//2,ns):
            d,j = sortlist[i]
            if d > foeval2 or i >= ns-1:  ##  variable in size (never null)
                paradat.outcasts[item].add(j)
            if i >= (ns-paradat.friends1):  ##  fixed size == friends1
                paradat.foes[item].add(j)

    return  ##  paradat updated


##  GRID2DIS :
def grid2dis (paradat,grid):
    "converts data grid to distance/dissimilarity matrix, as dictionary."

    ddic = {}
    ##  fill distable :
    vexp = 1
    if paradat.distmode != "city":
        vexp = 2
    for i,r in enumerate(grid):
        for j in range(paradat.rows):
            pair = (i,j)
            if i == j:
                ddic[pair] = 0.0
            elif j < i:
                ddic[pair] = ddic[(j,i)]
            else:
                xvec = [r[g] for g in paradat.goodlist]
                yvec = [grid[j][g] for g in paradat.goodlist]
                d = vecdist(xvec,yvec,vexp)
                ddic[pair] = d

    return  ddic        


##  DIS2DIC :
def dis2dic (grid):
    "makes distance grid into dictionary."

    ddic = {}
    rows = len(grid)  ##  known to be square, so cols==rows
    for r in range(rows):
        for c in range(rows):  ##  complete coverage
            pair = (r,c)
            ddic[pair] = grid[r][c]

    return  ddic


##  MAT2GRID :
def mat2grid (paradat):
    "makes grid from data matrix, scaling if required."

    ##  first compute basic stats :
    statlist = []
    print(NL+"variables to be used :")
    for c in range(paradat.cols):
        if c in paradat.goodcols:
            xvec = [paradat.datamat[r][c] for r in range(paradat.rows)]
            bs = bastatz(xvec)
            bs.xmin = min(xvec) ; bs.xmax = max(xvec)
            bs.xr = (bs.xmax - bs.xmin) or 1.0
            if bs.std2 < 1e-12:
                bs.std2 = 1.0
        else:
            bs = None
        statlist.append(bs)
        if bs:
            print(c,paradat.taglist[c])
            showobj(bs)
    ##  now do scaling :
    grid = []
    for r in range(paradat.rows):
        rowdata = []
        for c,bs in enumerate(statlist):
            x = paradat.datamat[r][c]
            if not bs:  ##  unused column
                x = 0.0
            elif paradat.scaling < 1:
                pass
            elif paradat.scaling == 1:
                x = (x - statlist[c].xmin) / statlist[c].xr
            else:
                x = (x - statlist[c].mean) / statlist[c].std2
            rowdata.append(x)
        grid.append(rowdata)

    return  grid  ##  scaled version of datamat (still with null cols).


##  INIT :
def init (paradat):
    "sets up data matrices."

    ##  read from data file if indicated :
    if paradat.datfile and os.path.exists(paradat.datfile):  ##  read x,y data
        r = getfile(paradat)
        if r < 0:
            raise RuntimeError("Trouble with input file !")
        else:  ##  data read in okay
            paradat.disgiven = distest(paradat.datamat)
            if paradat.disgiven:  ##  distances given as input
                paradat.ddic = dis2dic(paradat.datamat)
            else:  ##  convert rectangular grid to distances
                paradat.grid = mat2grid(paradat)
                paradat.ddic = grid2dis(paradat,paradat.grid)
    else:
        raise RuntimeError("Can't find datfile : "+paradat.datfile)
    
    ##  test cases :
    rows = paradat.rows
    m = rows // 2
    for i in []:
        for j in [1,m,rows-2]:
            print(i,j)
            if not paradat.disgiven:
                print(paradat.grid[i])
                print(paradat.grid[j])
            print(paradat.ddic[(i,j)])
            
    ##  raise RuntimeError("They shall not pass!!")

    ##  fill distance vector :
    paradat.dvec = []
    paradat.nt = nt = rows
    for t in range(nt):
        for tt in range(t+1,nt):
            pair = (t,tt)
            paradat.dvec.append(paradat.ddic[pair])
    ##  compute distances versus median :
    paradat.distomed = {}
    paradat.dmid = midvalue(paradat.dvec)
    paradat.distote = sum(paradat.dvec)
    paradat.dvec.sort()
    ##  lo = min(paradat.dvec) ; hi = max(paradat.dvec)
    ##  find nearest & furthest of each item :
    paradat.nearest,paradat.furthest = nearfar(paradat)
    assert len(paradat.nearest) == len(paradat.furthest)
    print(len(paradat.nearest),"buddies.",file=paradat.listfile)
    print("buddies :",paradat.nearest,file=paradat.listfile)
    nd = len(paradat.dvec)
    paradat.dbar = paradat.distote / nd
    p1 = nd // 4
    p2 = int(nd*0.75)
    lo = paradat.dvec[p1] ; hi = paradat.dvec[p2]
    ##  fill distance-vs-median(dist) table :
    for t in range(nt):
        for tt in range(nt):
            pair = (t,tt)
            d = paradat.ddic[pair]
            if d > hi:
                paradat.distomed[pair] = 2
            elif d <= lo:
                paradat.distomed[pair] = -2
            elif d > paradat.dmid:
                paradat.distomed[pair] = 1
            elif d < paradat.dmid:
                paradat.distomed[pair] = -1
            else:
                paradat.distomed[pair] = 0.0
            ##  adjust according to individual extremes :
            ##  if tt == paradat.furthest[t]:
                ##  paradat.distomed[pair] = 4
            if tt == paradat.nearest[t]:
                paradat.distomed[pair] = -4
    ##  identify near & distant subsets for each item :
    pickpals(paradat)
    print("friends1,friends2 =",paradat.friends1,paradat.friends2,file=paradat.listfile)
    paradat.nfriends = [paradat.friends1] * paradat.rows
    for t in range(0,nt,100):  ##  dump some debugging info :
        print(t,sorted(paradat.pals[t]),file=paradat.listfile)
        print(sorted(paradat.comrades[t]),file=paradat.listfile)
        print(len(paradat.pals[t]),len(paradat.foes[t]),file=paradat.listfile)
        print(len(paradat.comrades[t]),len(paradat.outcasts[t]),
              file=paradat.listfile)

    paradat.dmin = min(paradat.dvec)
    paradat.dmax = max(paradat.dvec)
    d2 = [d**2 for d in paradat.dvec]
    paradat.dsqtotal = sum(d2)
    paradat.dvar = paradat.dsqtotal / nd
    paradat.stopdist = paradat.dmid / nt
    paradat.basedist = paradat.dmid * Gold
              
    return  paradat
    

##  DUMPOUT :
def dumpout (dumpfile,paradat,bestitem,dlim=Htab):
    "writes extended data file to dumpfile, delimited with headline."

    headline = dlim.join(paradat.taglist)
    headline += dlim + "clus_ter"  ##  last column is group code
    print(headline,file=dumpfile)

    nd = len(paradat.datamat)
    ##  dump with extra group-code field :
    for j in range(nd):
        datline = paradat.datamat[j]  ##  input data case
        datline.append(bestitem[j])  ##  extra field
        datline = [str(d) for d in datline]
        datline = dlim.join(datline)
        print(datline,file=dumpfile)       

    return


##  POSTTIDY :
def posttidy (paradat,bestclus,bestval,ff):
    "tries to improve final clustering by greedy 1-step search."

    clusnums = sorted(set(bestclus))
    ng = len(clusnums)
    glen = len(bestclus)
    newc = bestclus[:]
    pointers = list(range(glen))
    paradat.included = set(pointers)  ##  just in case
    paradat.inclist = pointers[:]
    looping = True ; loop = 0
    if paradat.rows*ng > 25000:  ##  ng > 25 ??
        divisor = 2.0
    else:
        divisor = Gold
    ##  mx = int(paradat.rows / Gold)
    looptop = max(2,10 - paradat.ncycles)
    mx = glen  ##  items to consider
    while looping and loop < looptop:
        if loop < 1:  ##  first time
            random.shuffle(pointers)  ##  just pick at random
        else:  ##  subsequent loops
            mislist = misfits(newc,paradat)  ##  worse fits towards front
            ##  print(mislist[:7]) ; print(mislist[-1])
            pointers = [mislist[j][-1] for j in range(mx)]
        print(mx)
        mx = int(mx / divisor)  ##  ready for next time
        looping = False ; loop += 1
        print("post-tidying loop",loop,"begins with best score",round(bestval,7))
        for j,p in enumerate(pointers):
            bc = newc[p]
            for c in clusnums:  ##  test each possible group
                newc[p] = c
                f = ff(newc,paradat)
                if f > bestval:  ##  improvement
                    ##  print("!!",loop,j,p,bc,c,round(f,7))
                    bc = c ; bestval = f
                    looping = True
            newc[p] = bc
        print("post-tidying loop",loop,"ends with best score",round(bestval,7))

    return  newc, bestval


##  REUNITE :
def reunite (paradat,bestclus,bestval,ff):
    "reunites neighbours if it would improve overall score."

    clusnums = sorted(set(bestclus))
    ng = len(clusnums)
    glen = len(bestclus)
    newc = bestclus[:]
    ob = bestval
    pointers = list(range(glen))
    paradat.included = set(pointers)  ##  just in case
    paradat.inclist = pointers[:]
    ##  scan thru seeking improvement :
    for loop in [1,2]:
        random.shuffle(pointers)
        for j,p in enumerate(pointers):
            bc = newc[p]
            nn = paradat.nearest[p]
            nc = newc[nn]
            if bc != nc:  ##  in different clusters
                newc[p] = nc
                f = ff(newc,paradat)
                if f > bestval:  ##  improvement
                    ##  print('!',loop,j,p,bc,nc,round(f,7))
                    bc = nc ; bestval = f
            newc[p] = bc
    if ob < bestval:
        print(':',round(bestval,7))
        
    return  newc,bestval




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
